package main

import (
	"crypto"
	"encoding/hex"
	"encoding/json"
	"strconv"
	"strings"
	"ahe/PSW/api/ChainCode"
	"github.com/hyperledger/fabric/core/chaincode/shim"
	pb "github.com/hyperledger/fabric/protos/peer"
)

var logger = shim.NewLogger("TransChaincode Demo")

var NEWBALANCE_A = "newBalanceA"
var NEWBALANCE_B = "newBalanceB"
var NEWTRANSACTION_A = "newTxA"
var NEWTRANSACTION_B = "newTxB"
var TX_PAY = "P"
var TX_RECEIVE = "R"
var TX_INIT = "I"

type TransChaincodeDemo struct{}

type TransRecord struct {
	FromAddr string
	ToAddr   string
	TXType   string
	Balance  string
	TX       string
	remark   string
}

func (t *TransChaincodeDemo) Init(stub shim.ChaincodeStubInterface) pb.Response {
	return shim.Success(nil)
}

func (t *TransChaincodeDemo) Invoke(stub shim.ChaincodeStubInterface) pb.Response {
	function, args := stub.GetFunctionAndParameters()

	if function == "QueryBalance" {
		return t.queryBalance(stub, args)
	} else if function == "Transfer" {
		return t.transfer(stub, args)
	} else if function == "init" {
		return t.init(stub, args)
	} else if function == "HomoAdd" {
		return t.homoAdd(stub, args)
	} else if function == "HomoMulti" {
		return t.homoMulti(stub, args)
	}

	return shim.Error("Invalid invoke function name: " + function)
}

func (t *TransChaincodeDemo) transfer(stub shim.ChaincodeStubInterface, args []string) pb.Response {

	if len(args) != 3 {
		logger.Error("Incorrect number of arguments. expect 3 arguments")
		return shim.Error("Incorrect number of arguments. expect 3 arguments")
	}

	AddrA := args[0]
	AddrB := args[1]
	txInfo := args[2]

	if strings.Compare(AddrA, AddrB) == 0 {
		logger.Error("A' addr is the same B'Addr")
		return shim.Error("A' addr is the same B'Addr")
	}

	// read a's trans record
	transRecA, err := stub.GetState(AddrA)
	if err != nil {
		return shim.Error("Failed to get state")
	}
	if transRecA == nil {
		return shim.Error("Entity not found")
	}

	var transRecAStruct = TransRecord{}
	err = json.Unmarshal(transRecA, &transRecAStruct)
	if err != nil {
		logger.Error("fail to unmarshal user's trans record")
		return shim.Error("fail to unmarshal user's trans record")
	}

	// read b's trans record
	transRecB, err := stub.GetState(AddrB)
	if err != nil {
		return shim.Error("Failed to get state")
	}
	if transRecA == nil {
		return shim.Error("Entity not found")
	}

	var transRecBStruct = TransRecord{}
	err = json.Unmarshal(transRecB, &transRecBStruct)
	if err != nil {
		logger.Error("fail to unmarshal user's trans record")
		return shim.Error("fail to unmarshal user's trans record")
	}

	//validate transfer information
	cipherBalanceAKeyABlock := transRecAStruct.Balance
	cipherBalanceBKeyBBlock := transRecBStruct.Balance

	newCipherBalanceA, newCipherBalanceB, newCipherTxA, newCipherTxB, err := pswapi_cc.ValidateTxInfo(txInfo, cipherBalanceAKeyABlock, cipherBalanceBKeyBBlock)
	if err != nil {
		logger.Error("fail to validate trans information")
		return shim.Error("fail to validate trans information")
	}

	// update a's balance
	transRecAStruct.Balance = newCipherBalanceA
	transRecAStruct.TX = newCipherTxA
	transRecAStruct.TXType = "P"

	AvalbytesUpdate, err := json.Marshal(transRecAStruct)
	if err != nil {
		logger.Error("fail to marshal balance update info")
		return shim.Error("Marshal Error")
	}

	err = stub.PutState(AddrA, AvalbytesUpdate)
	if err != nil {
		logger.Error("fail to store state: ", err.Error())
		return shim.Error(err.Error())
	}

	// update b's balance
	transRecBStruct.Balance = newCipherBalanceB
	transRecBStruct.TX = newCipherTxB
	transRecBStruct.TXType = "R"
	BvalbytesUpdate, err := json.Marshal(transRecBStruct)
	if err != nil {
		logger.Error("fail to marshal balance update info")
		return shim.Error("Marshal Error")
	}

	err = stub.PutState(AddrB, BvalbytesUpdate)
	if err != nil {
		return shim.Error(err.Error())
	}

	return shim.Success([]byte("Success"))
}

func (t *TransChaincodeDemo) init(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	if len(args) != 2 {
		logger.Error("parameters number is not correct")
		return shim.Error("parameters number is not correct")
	}

	PubKey := string(args[0])
	BalanceInfo := string(args[1])

	hashPubkey, err := t.calcAddr(PubKey)

	UserPubKey, err := stub.GetState(hashPubkey)
	if err != nil {
		logger.Error("Error on query addr")
		return shim.Error("fail to query addr")
	}

	if UserPubKey != nil {
		logger.Error("addr already register")
		return shim.Error("addr already register")
	}

	//validate balance
	cipherBalance, err := pswapi_cc.ValidateInitBalance(BalanceInfo, PubKey)
	if err != nil {
		logger.Error("fail to Validate InitBalance ")
		return shim.Error("fail toValidate InitBalance")
	}

	TxRec := &TransRecord{}
	TxRec.FromAddr = "sys"
	TxRec.ToAddr = hashPubkey
	TxRec.Balance = string(cipherBalance)
	TxRec.TX = string(cipherBalance)
	TxRec.TXType = TX_INIT

	TxRecStore, err := json.Marshal(TxRec)
	if err != nil {
		logger.Error("fail to marshal trans record")
		return shim.Error("fail to marshal trans record")
	}

	// store user's trans record
	err = stub.PutState(hashPubkey, []byte(TxRecStore))
	if err != nil {
		logger.Error("fail to store trans record")
		return shim.Error("fail to store trans record")
	}

	return shim.Success([]byte("Success"))
}

// query account's balance
func (t *TransChaincodeDemo) queryBalance(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	if len(args) != 1 {
		logger.Error("Incorrect number of arguments. Expecting addr to query")
		return shim.Error("Incorrect number of arguments. Expecting addr to query")
	}

	Addr := string(args[0])
	// Get the state from the ledger
	balance, err := stub.GetState(Addr)
	if err != nil {
		logger.Error("fail to get state for: ", Addr)
		return shim.Error("fail to get state for: " + Addr)
	}

	if balance == nil {
		logger.Error("nil amount for: ", Addr)
		return shim.Error("nil amount for: " + Addr)
	}

	return shim.Success(balance)
}

// call homomorphic addition function
func (t *TransChaincodeDemo) homoAdd(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	if len(args) != 2 {
		logger.Error("Incorrect number of arguments. Expecting two cipher")
		return shim.Error("Incorrect number of arguments. Expecting two cipher")
	}

	cipher1 := string(args[0])
	cipher2 := string(args[1])
	// Get the state from the ledger
	homoaddRes, err := pswapi_cc.Add(cipher1, cipher2)
	if err != nil {
		logger.Error("error on homoadd", err.Error())
		return shim.Error("error on homoadd" + err.Error())
	}

	return shim.Success([]byte(homoaddRes))
}

func (t *TransChaincodeDemo) homoMulti(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	if len(args) != 2 {
		logger.Error("Incorrect number of arguments. Expecting two cipher")
		return shim.Error("Incorrect number of arguments. Expecting two cipher")
	}

	cipher1 := string(args[0])
	times := string(args[1])
	timesInt, err := strconv.Atoi(times)
	if err != nil {
		logger.Error("error on homomulti", err.Error())
		return shim.Error("error on parse times" + err.Error())
	}
	homomultiRes := cipher1
	// Get the state from the ledger
	for i := 1; i < timesInt; i++ {
		homomultiResTemp, err := pswapi_cc.Add(cipher1, homomultiRes)
		homomultiRes = homomultiResTemp
		if err != nil {
			logger.Error("error on homomulti", err.Error())
			return shim.Error("error on homomulti" + err.Error())
		}
	}

	return shim.Success([]byte(homomultiRes))
}

func (t *TransChaincodeDemo) calcAddr(cont string) (string, error) {

	Hasher := crypto.SHA256.New()
	Hasher.Write([]byte(cont))
	HashRes := Hasher.Sum(nil)

	return hex.EncodeToString(HashRes), nil
}

func main() {
	err := shim.Start(new(TransChaincodeDemo))
	if err != nil {
		logger.Error("Error starting Simple chaincode: ", err.Error())
	}
}
